# Copyright (c) 2024, InfinityQ Technology, Inc.
import logging
from typing import Any, Optional
import numpy as np
import numpy.typing as npt

from .base_constraints import BaseConstraints
from ...errors import ConstraintSizeError, MaximumConstraintLimitError

log = logging.getLogger("TitanQ")

_MAX_CONSTRAINTS_COUNT = 32_000


class Constraints(BaseConstraints):
    def __init__(self) -> None:
        self._constraint_weights = None
        self._constraint_bounds = None


    def is_empty(self) -> bool:
        """return if all constraints are empty"""
        return self._constraint_weights is None and self._constraint_bounds is None


    def add_constraint(
        self,
        variable_size: int,
        constraint_weights: npt.NDArray[Any],
        constraint_bounds: npt.NDArray[Any]
    ) -> None:
        """
        Add a constraint to the existing ones

        :param variable_size: the variable size from the model
        :param constraint_weights: constraint_weights to append to the existing ones.
        :param constraint_bounds: constraint_bounds to append to the existing ones.

        :raises ConstraintSizeError: constraint size is different than variable size.
        :raises MaximumConstraintLimitError: the number of constraint exeed the limit.
        """
        # shape validation
        if constraint_weights.shape[1] != variable_size:
            raise ConstraintSizeError(
                "Constraint mask shape does not match the variable size. " \
                f"Constraint size: {constraint_weights.shape[1]}, Variable size: {variable_size}")

        # limit validation
        if variable_size + self._rows_count(self._constraint_weights) + constraint_weights.shape[0] > _MAX_CONSTRAINTS_COUNT:
            raise MaximumConstraintLimitError(
                "Cannot add additional constraints. The limit of constraints have been reached. " \
                f"The sum of variables and constraints cannot exceed {_MAX_CONSTRAINTS_COUNT}." \
                f"Number of constraints: {self._rows_count(self._constraint_weights)}; " \
                f"Number of variables: {variable_size}; " \
                f"while trying to add {constraint_weights.shape[0]} new constraints."
            )

        self._constraint_weights = self._append_constraint(constraint_weights, self._constraint_weights)
        self._constraint_bounds = self._append_constraint(constraint_bounds, self._constraint_bounds)


    def weights(self) -> Optional[npt.NDArray[np.float32]]:
        """
        :return: The weights constraints.
        """
        return self._constraint_weights


    def bounds(self) -> Optional[npt.NDArray[np.float32]]:
        """
        :return: The bounds constraints.
        """
        return self._constraint_bounds



# Copyright (c) 2024, InfinityQ Technology, Inc.
from abc import abstractmethod, ABC
import numpy as np
import numpy.typing as npt

from ..numpy_util import convert_to_float32


class BaseConstraints(ABC):

    @abstractmethod
    def is_empty():
        """Return if all constraints are empty"""
        pass


    def _append_constraint(
        self,
        new_constraints: npt.NDArray[np.float32],
        existing_constraints: npt.NDArray[np.float32]
    ) -> npt.NDArray[np.float32]:
        """
        Appends new_constraints to existing_constraints and return the new formed one

        :param new_constraints: New constraint to append with the existing_constraints
        :param existing_constraints: Existing constraint that will be appended with

        :return: Existing constraint with the appeneded new constraints
        """
        # API only take float 32
        new_constraints = convert_to_float32(new_constraints)

        if existing_constraints is None:
            return new_constraints
        else:
            return np.append(existing_constraints, new_constraints, axis=0)


    def _rows_count(self, constraints: npt.NDArray[np.float32]) -> int:
        """
        :return: The number of constraints (rows) already set, 0 if never set
        """
        if constraints is None:
            return 0
        return constraints.shape[0]


# Copyright (c) 2024, InfinityQ Technology, Inc.
import logging
from typing import Any, Optional
import numpy as np
import numpy.typing as npt

from .base_constraints import BaseConstraints
from ...errors import ConstraintSizeError, MaximumConstraintLimitError

log = logging.getLogger("TitanQ")

_MAX_QUAD_CONSTRAINTS_COUNT = 8


class QuadConstraints(BaseConstraints):
    def __init__(self) -> None:
        self._quad_constraint_weights = None
        self._quad_constraint_bounds = None
        self._quad_constraint_linear_weights = None


    def is_empty(self) -> bool:
        """return if all constraints are empty"""
        return (
            self._quad_constraint_weights is None and
            self._quad_constraint_bounds is None and
            self._quad_constraint_linear_weights is None
        )


    def add_constraint(
        self,
        variable_size: int,
        quad_constraint_weights: npt.NDArray[Any],
        quad_constraint_bounds: npt.NDArray[Any],
        quad_constraint_linear_weights: Optional[npt.NDArray[Any]] = None
    ) -> None:
        """
        Add a quadratic constraint to the existing ones

        :param variable_size: the variable size from the model
        :param quad_constraint_weights: quadratic constraint weights to append to the existing ones.
        :param quad_constraint_bounds: quadratic constraint bounds to append to the existing ones.
        :param quad_constraint_linear_weights: quadratic constraint linear weights to append to the existing ones.

        :raises ConstraintSizeError: constraint size is different than variable size.
        :raises MaximumConstraintLimitError: the number of constraint exeed the limit.
        """
        # shape validation
        if quad_constraint_weights.shape != (variable_size, variable_size):
            raise ConstraintSizeError(
                 "Invalid constraint_mask shape: expected NxN where N is the number of variables " \
                f"({variable_size}), but got {quad_constraint_weights.shape}.")

        if quad_constraint_linear_weights is not None and quad_constraint_linear_weights.shape != (1, variable_size):
                raise ValueError(
                "Invalid constraint_linear_weights shape: expected (N,) where N is " \
                f"the number of variables ({variable_size}), " \
                f"but got {quad_constraint_linear_weights.shape}.")

        # limit validation
        if self._rows_count(self._quad_constraint_weights) > (_MAX_QUAD_CONSTRAINTS_COUNT * variable_size):
            raise MaximumConstraintLimitError(
                "Cannot add additional quadratic constraints. " \
                f"The limit of quadratic constraints ({_MAX_QUAD_CONSTRAINTS_COUNT}) has been reached."
            )

        self._quad_constraint_weights = self._append_constraint(quad_constraint_weights, self._quad_constraint_weights)
        self._quad_constraint_bounds = self._append_constraint(quad_constraint_bounds, self._quad_constraint_bounds)

        if quad_constraint_linear_weights is not None:
            self._quad_constraint_linear_weights = self._append_constraint(
                quad_constraint_linear_weights,
                self._quad_constraint_linear_weights)


    def weights(self) -> Optional[npt.NDArray[np.float32]]:
        """
        :return: The quadratic weights constraints.
        """
        return self._quad_constraint_weights


    def bounds(self) -> Optional[npt.NDArray[np.float32]]:
        """
        :return: The quadratic bounds constraints.
        """
        return self._quad_constraint_bounds

    def linear_weights(self) -> Optional[npt.NDArray[np.float32]]:
        """
        :return: The quadratic linear weights constraints.
        """
        return self._quad_constraint_linear_weights



# Copyright (c) 2024, InfinityQ Technology, Inc.
from typing import List

from .model import MPSObject
from .options import MPSParseOptions
from ._validations import (
    validate_missing_required_sections,
    validate_bounds_identifiers,
    validate_columns_identifiers,
    validate_ranges_identifiers,
    validate_rhs_identifiers
)
from ._visitor import LineSection
from ._section_parser import *
from ._section_reader import SectionReader


def parse_from_mps(path: str, options: MPSParseOptions) -> MPSObject:
    """
    import an .mps file and parse it as an MPSObject (python object)

    :param path: .mps file path
    :param options: MPSOptions if needed, default will be set if not provided

    :returns: A python like object of the .mps
    :rtype: MPSObject

    Example of this function:

    .. code-block:: python
    >>> mps_object = parse_from_mps("PATH_TO_FILE")
    >>> # name of the problem
    >>> name = mps_object.name
    >>> # first row sense
    >>> first_row = mps_object.rows[0].sense()
    """
    section_reader = SectionReader(path, options)
    parsed_lines_section: List[LineSection] = []

    # generator
    for section in section_reader.sections():
        parsed_lines_section.extend(section.parse())

    mps_object = MPSObject()
    for line in parsed_lines_section:
        line._accept(mps_object)

    # post-validations
    validate_missing_required_sections(mps_object)
    validate_columns_identifiers(mps_object.rows, mps_object.free_row, mps_object.columns)
    validate_rhs_identifiers(mps_object.rows, mps_object.rhs)
    validate_ranges_identifiers(mps_object.rows, mps_object.ranges)
    validate_bounds_identifiers(mps_object.columns, mps_object.bounds)

    return mps_object


def export_to_mps(mps_object: MPSObject, path: str) -> None:
    raise NotImplementedError("Exporting to MPS is not currently supported")


# Copyright (c) 2024, InfinityQ Technology, Inc.
from typing import Any, List, Tuple

from ..mps_parser import ColumnsType, MPSColumns


class SetOnceOrDefault:
    def __init__(
        self,
        default: Any,
        error_msg: str,
        allow_same_value: bool = False
    ) -> None:
        """
        SetOnceOrDefault creates an object with a default value that can only
        be set once. If not set, will use the default value.

        If allow_same_value is set to True, the object can be set again but has
        to be the exact same value.
        """
        self._default = default
        self._error_msg = error_msg
        self._allow_same_value = allow_same_value

        self._value = None

    def is_default(self) -> bool:
        """returns if the value is still at default state"""
        return self._value is None

    def set(self, value: Any):
        # a value has already been set
        if self._value is not None:
            # check if we allow the same value
            if self._allow_same_value and self._value == value:
                pass
            else:
                raise Exception(
                    f"{self._error_msg}: Tried to set twice a {self.__class__.__name__} " +
                    f"and had already the value '{self._value}' set, but tried to set it again with '{value}'")
        self._value = value

    def get(self) -> Any:
        return self._value if self._value is not None else self._default


def get_variables_and_types(mps_columns: List[MPSColumns]) -> List[Tuple[str, ColumnsType]]:
    """returns a list without duplicates but keeps an order"""
    variables_set = []
    seen = set()
    seen_add = seen.add # more efficient

    for column in mps_columns:
        if column.identifier() not in seen:
            seen_add(column.identifier())
            variables_set.append((column.identifier() ,column.type()))

    return variables_set



# Copyright (c) 2024, InfinityQ Technology, Inc.
from .base_constraints import BaseConstraints
from .constraint_type import ConstraintType, get_constraint_type
from .constraints import Constraints
from .quad_constraints import QuadConstraints


# Copyright (c) 2024, InfinityQ Technology, Inc.
import abc
from warnings import warn
from pyparsing import LineEnd, Opt, Or, ParseException, ParserElement, alphanums, nums, Group, OneOrMore, Word, printables
from typing import List, Union

from titanq.errors import MpsMalformedFileError, MpsMissingValueError

from .model import (
    MPSBounds,
    MPSColumns,
    MPSEndata,
    MPSName,
    MPSQuadobj,
    MPSRanges,
    MPSRhs,
    MPSRow,
)
from .utils import BoundsType, ColumnsType, RowsType, SectionType, UniqueList
from ._visitor import LineSection


# Following this IBM documentation, this parser currently uses spaces as delimiters
# https://www.ibm.com/docs/en/icos/22.1.0?topic=standard-records-in-mps-format

# For integer variables both ways are supported (COLUMNS via markers and in the bounds section)
# https://www.ibm.com/docs/en/icos/22.1.1?topic=extensions-integer-variables-in-mps-files

# mps char limits
_MAX_IDENTIFIER_LENGTH = 255
_MAX_NUMERIC_LENGTH = 25

# Integer variables markers in COLUMNS
_INTEGER_MARKER = "'MARKER'" # 'MARKER'
_INTEGER_START = "'INTORG'" # 'INTORG'
_INTEGER_END = "'INTEND'" # 'INTEND'

# Pyparser grammar
number = Word(nums + "-.", max=_MAX_NUMERIC_LENGTH) # numbers can have decimals and can be negative
rows_type = Or([type.value for type in RowsType])
bounds_type = Or([type.value for type in BoundsType])
identifier = Word(alphanums, max=_MAX_IDENTIFIER_LENGTH)


def _parse_with_error_handling(element: ParserElement, line: str, line_index: str):
    """ Wrap ParseException with MPSMalformedFileError"""
    try:
        return element.parseString(line)
    except ParseException as ex:
        # As an example, this should read as:
        # "Expected {'N' ^ 'G' ^ 'L' ^ 'E'}, found 'AG  ROW01' (line: 4)"
        raise MpsMalformedFileError(f"{ex.msg}, found '{ex.pstr}' (line: {line_index})")


class SectionParser(abc.ABC):
    """class interface for any section needed to be parsed"""
    def __init__(self, start_index: int, lines: List[str]) -> None:
        super().__init__()
        self._start_index = start_index
        self._lines = lines

    def parse() -> List[LineSection]:
        """parse the given lines"""


class NameParser(SectionParser):
    """parses the NAME section of an .mps file into python objects"""

    def parse(self) -> List[MPSName]:
        # possible format --> ['NAME', 'TITANQ-1-Project-50%']
        if len(self._lines) > 1:
            raise MpsMalformedFileError(f"Found more than 2 lines for section '{SectionType.NAME}'")

        element = SectionType.NAME.value + Word(printables, max=_MAX_IDENTIFIER_LENGTH) + LineEnd()
        parsed_elements = element.parseString(self._lines[0])
        return [MPSName(name=parsed_elements[1])]


class RowsParser(SectionParser):
    """parses the ROWS section of an .mps file into python objects"""

    def __init__(self, start_index: int, lines: List[str]) -> None:
        super().__init__(start_index, lines)
        self._rows_identifier = UniqueList(f"Found duplicates in the '{SectionType.ROWS}' section")
        self._free_row = None


    def _set_free_row(self, free_row_identifier):
        if self._free_row is not None:
            warn(
                f"Found a free row {RowsType.FREE_ROW} identified as '{free_row_identifier}', "
                f"but one was already identified with: '{self._free_row}'. New one will be ignored"
            )
        self._free_row = free_row_identifier


    def parse(self) -> List[MPSRow]:
        # possible format --> ['N', 'OBJ']
        rows = []
        for index, line in enumerate(self._lines):
            element = rows_type + identifier + LineEnd()
            parsed_elements = _parse_with_error_handling(element, line, self._start_index + index)

            row_identifier = parsed_elements[1]
            sense = RowsType(parsed_elements[0])
            if sense == RowsType.FREE_ROW:
                self._set_free_row(row_identifier)

            rows.append(MPSRow(sense=sense, identifier=row_identifier))

        if self._free_row is None:
            raise MpsMissingValueError(f"Did not find a free row '{RowsType.FREE_ROW}' in the '{SectionType.ROWS}' section")

        return rows


class ColumnParser(SectionParser):
    """parses the COLUMNS section of an .mps file into python objects"""

    def __init__(self, start_index: int, lines: List[str]) -> None:
        super().__init__(start_index, lines)
        self._type = ColumnsType.CONTINUOUS


    def _activate_integer_mode(self): self._type = ColumnsType.INTEGER
    def _deactivate_integer_mode(self): self._type = ColumnsType.CONTINUOUS


    def _handle_marker_line(self, value: str) -> None:
        if value == _INTEGER_START:
            self._activate_integer_mode()
        elif value == _INTEGER_END:
            self._deactivate_integer_mode()


    def parse(self) -> List[MPSColumns]:
        # possible format 1 --> ['C1', ['R1', '2.0']]
        # possible format 2 --> ['C2', ['R2', '-2.0'], ['R3', '-1.0]]
        # possible format 3 --> ['MARK', 'MARKER, 'INTORG']
        # possible format 4 --> ['SOMENAME', 'MARKER, 'INTEND']
        columns = []
        for index, line in enumerate(self._lines):
            row_element = Group(Or([identifier, _INTEGER_MARKER]) + Or([number, _INTEGER_START, _INTEGER_END]))
            element = identifier + OneOrMore(row_element)
            parsed_elements = _parse_with_error_handling(element, line, self._start_index + index)

            row_idendtifier = parsed_elements[1][0]

            if row_idendtifier == _INTEGER_MARKER:
                self._handle_marker_line(parsed_elements[1][1])
                continue # skip this line

            columns.append(MPSColumns(
                identifier=parsed_elements[0],
                row_identifier=row_idendtifier,
                coeff=float(parsed_elements[1][1]),
                type=self._type
            ))

            # this means we have an additional column on the same line
            if len(parsed_elements) > 2:
                columns.append(MPSColumns(
                    identifier=parsed_elements[0],
                    row_identifier=parsed_elements[2][0],
                    coeff=float(parsed_elements[2][1]),
                    type=self._type
                ))

            if len(parsed_elements) > 3:
                raise MpsMalformedFileError(f"Found more than 2 rows identifier in a single '{SectionType.COLUMNS}' line")

        return columns


class RhsParser(SectionParser):
    """parses the RHS section of an .mps file into python objects"""

    def parse(self) -> List[MPSRhs]:
        # possible format 1 --> ['RHS1', ['R1', '2.0']]
        # possible format 2 --> ['RHS2', ['R2', '-2.0'], ['R3', '-1.0]]
        rhs = []
        for index, line in enumerate(self._lines):
            row_element = Group(identifier + number)
            element = identifier + OneOrMore(row_element)
            parsed_elements = _parse_with_error_handling(element, line, self._start_index + index)

            rhs.append(MPSRhs(
                identifier=parsed_elements[0],
                row_identifier=parsed_elements[1][0],
                coeff=float(parsed_elements[1][1])
            ))

            # this means we have an additional rhs on the same line
            if len(parsed_elements) > 2:
                rhs.append(MPSRhs(
                    identifier=parsed_elements[0],
                    row_identifier=parsed_elements[2][0],
                    coeff=float(parsed_elements[2][1])
                ))

            if len(parsed_elements) > 3:
                raise MpsMalformedFileError(f"Found more than 2 rows identifier in a single '{SectionType.RHS}' line")

        return rhs


class RangesParser(SectionParser):
    """parses the RANGES section of an .mps file into python objects"""

    def parse(self) -> List[MPSRanges]:
        # possible format 1 --> ['RNG1', ['R1', '2.0']]
        # possible format 2 --> ['RNG2', ['R2', '-2.0'], ['R3', '-1.0]]
        ranges = []
        for index, line in enumerate(self._lines):
            row_element = Group(identifier + number)
            element = identifier + OneOrMore(row_element)
            parsed_elements = _parse_with_error_handling(element, line, self._start_index + index)

            ranges.append(MPSRanges(
                identifier=parsed_elements[0],
                row_identifier=parsed_elements[1][0],
                coeff=float(parsed_elements[1][1]),
            ))

            # this means we have an additional range on the same line
            if len(parsed_elements) > 2:
                ranges.append(MPSRanges(
                    identifier=parsed_elements[0],
                    row_identifier=parsed_elements[2][0],
                    coeff=float(parsed_elements[2][1]),
                ))

            if len(parsed_elements) > 3:
                raise MpsMalformedFileError(f"Found more than 2 rows identifier in a single '{SectionType.RANGES}' line")

        return ranges


class BoundParser(SectionParser):
    """parses the BOUNDS section of an .mps file into python objects"""

    def parse(self) -> List[MPSBounds]:
        # possible format 1 --> ['UP', 'BND1', 'COL01', '1.0' ]
        # possible format 2 --> ['FR', 'BND2', 'COL02' ]
        bounds = []
        for index, line in enumerate(self._lines):
            element = bounds_type + identifier + identifier + Opt(number) + LineEnd()
            parsed_elements = _parse_with_error_handling(element, line, self._start_index + index)

            bounds.append(MPSBounds(
                identifier=parsed_elements[1],
                type=BoundsType(parsed_elements[0]),
                column_identifier=parsed_elements[2],
                value=float(parsed_elements[3]) if len(parsed_elements) > 3 else None
            ))

        return bounds


class QuadobjParser(SectionParser):
    """parses the QUADOBJ or QMATRIX section of an .mps file into python objects"""

    def __init__(self, start_index: int, lines: List[str], type: Union[SectionType.QUADOBJ, SectionType.QMATRIX]) -> None:
        super().__init__(start_index, lines)
        self._type = type

    def parse(self) -> List[MPSQuadobj]:
        # possible format --> ['x', 'x', 3]
        # possible format --> ['x', 'y', 1]
        quadobj = []
        for index, line in enumerate(self._lines):
            element = identifier + identifier + number + LineEnd()
            parsed_elements = _parse_with_error_handling(element, line, self._start_index + index)

            row = parsed_elements[0]
            column = parsed_elements[1]
            value = float(parsed_elements[2])

            quadobj.append(MPSQuadobj(row_identifier=row, column_identifier=column, value=float(parsed_elements[2])))

            # if QUADOBJ is set instead of QMATRIX, write another line but inversed to fulfill the matrix
            if self._type == SectionType.QUADOBJ and row != column:
                quadobj.append(MPSQuadobj(row_identifier=column, column_identifier=row, value=value))

        return quadobj


class EndataParser(SectionParser):
    """parses the ENDATA section of an .mps file into a python objects"""

    def parse(self) -> List[MPSEndata]:
        # possible format --> ['ENDATA'] # no need to parse here
        if len(self._lines) > 1:
            raise MpsMalformedFileError(f"Found more than 2 lines for section '{SectionType.ENDATA}'")

        return [MPSEndata()]


# Copyright (c) 2024, InfinityQ Technology, Inc.
from typing import Any, Generator, Optional

from .options import MPSParseOptions
from ._section_parser import (
    BoundParser,
    ColumnParser,
    EndataParser,
    NameParser,
    QuadobjParser,
    RangesParser,
    RhsParser,
    RowsParser,
    SectionParser
)
from .utils import SectionType


class SectionReader:
    """
    Reads an MPS file and generate section by section with
    the method sections()
    """
    def __init__(self, path: str, options: MPSParseOptions) -> None:
        self._path = path
        self._options = options


    def _new_section(self, line: str) -> Optional[SectionType]:
        """returns a section if encountered a new one, returns None otherwise"""
        # an exception for NAME, where the value is on the same line
        if line.startswith(SectionType.NAME):
            return SectionType.NAME

        for section in [section.value for section in SectionType]:
            if line == section:
                return section
        return None


    def sections(self) -> Generator[SectionParser, Any, Any]:
        """generate sections by yielding one at a time"""
        current_section = None
        lines = []

        # open context manager is already lazy loading, since
        # the file is line-based. This way it loads into memory by chunks
        section_start_index = 1
        with open(self._path, 'r') as file:
            for line_index, line in enumerate(file, start=1):
                line = line.strip() # always strip the line
                if self._options.empty_line_check(line_index, line):
                    continue

                new_section = self._new_section(line)
                # this means we have entered a new section, we need to yield the old one
                if new_section is not None:
                    if current_section == SectionType.ROWS:
                        yield RowsParser(section_start_index, lines)
                    elif current_section == SectionType.COLUMNS:
                        yield ColumnParser(section_start_index, lines)
                    elif current_section == SectionType.RHS:
                        yield RhsParser(section_start_index, lines)
                    elif current_section == SectionType.RANGES:
                        yield RangesParser(section_start_index, lines)
                    elif current_section == SectionType.BOUNDS:
                        yield BoundParser(section_start_index, lines)
                    elif current_section == SectionType.QUADOBJ:
                        yield QuadobjParser(section_start_index, lines, SectionType.QUADOBJ)
                    # same as QUADOBJ, but QMATRIX specifies each entry of the matrix
                    elif current_section == SectionType.QMATRIX:
                        yield QuadobjParser(section_start_index, lines, SectionType.QMATRIX)

                    # NAME and ENDATA yield immediatly
                    if new_section == SectionType.NAME:
                        yield NameParser(section_start_index, [line])
                    elif new_section == SectionType.ENDATA:
                        yield EndataParser(section_start_index, [line])
                        break # end of file, stop here

                    lines = []
                    current_section = new_section
                    section_start_index = line_index + 1
                else:
                    lines.append(line) # just add the line



# Copyright (c) 2024, InfinityQ Technology, Inc.
import enum
from typing import Optional, Tuple

import numpy as np


class Target(enum.Enum):
    """
    All target types currently supported by the solver
    """

    MINIMIZE = 'minimize'


def _verify_nd_array(array: np.ndarray, array_name: str, *, expected_shape: Tuple, expected_type: np.dtype):
    """
    Make sure the given array is the right shape and the right type.

    :param array: The array to verify.
    :param array_name: the name of the array to be verified, use in error message.
    :param expected_shape: numpy compatible tuple of the expected shape.
    :param expected_type: expected numpy type for the array.

    :raise ValueError: if the array is not the right shape.
    :raise ValueError: if the value inside the array is not the right data type.
    """

    if array.shape != expected_shape:
        raise ValueError(f"{array_name} shape {array.shape} does not fit the shape of the variable previously defined. Expected: {expected_shape}.")

    if array.dtype != expected_type:
        raise ValueError(f"Unsupported {array_name} dtype ({array.dtype}). Expected: {expected_type}.")


class Objective:
    """
    Objective passed to TitanQ platform. It is consisted of the weight matrix and the bias vector.
    """

    def __init__(self, var_size: int, weights: Optional[np.ndarray], bias: np.ndarray, target: Target, constant_term: float) -> None:
        # weights matrix can be set to None, therefore we can skip the validations
        if weights is not None:
            _verify_nd_array(weights, "weights", expected_shape=(var_size, var_size), expected_type=np.float32)

        _verify_nd_array(bias, "bias", expected_shape=(var_size, ), expected_type=np.float32)

        self._weights = weights
        self._bias = bias
        self._target = target
        self._constant_term = constant_term

    def weights(self) -> Optional[np.ndarray]:
        """
        :return: The weights matrix of this objective.
        """
        return self._weights

    def bias(self) -> np.ndarray:
        """
        :return: The bias vector of this objective.
        """
        return self._bias

    def target(self) -> Target:
        """
        :return: The Target for this objective.
        """
        return self._target

    def constant_term(self) -> float:
        """
        :return: The constant term (offset) for this objective
        """
        return self._constant_term



# Copyright (c) 2024, InfinityQ Technology, Inc.

"""
Support for bipolar problems is deprecated in TitanQ.

This module can convert a bipolar problem into binary,
then the solution back from binary into bipolar.
This way, a bipolar problem can seamlessly be optimized by TitanQ
as a binary problem.
"""

import numpy as np

from typing import Optional, Tuple


class BipolarToBinary:
    """
    Convert a bipolar problem to be solved as binary.

    TitanQ doesn't support bipolar problems directly, but these can be converted
    to binary problems instead.

    The following steps are involved:

    1. Convert the weight matrix and bias vector.
    2. Create a :class:`titanq.Model` object with binary variables,
       using the converted weight matrix and bias vector.
    3. Optimize the model.
    4. Reuse the object created in step 1 to convert the optimization
       result (result vector and objective value).

    This class assists with steps 1 and 4.

    ⚠️ Warning
    ----------

    Ensure you convert the result back to bipolar after solving the problem.
    See :meth:`convert_result`.

    Example
    -------
    >>> # Step 1: Convert the bipolar problem.
    >>> # Let weights and bias define a BIPOLAR problem.
    >>> converter = BipolarToBinary(weights=weights, bias=bias, inplace=False)
    >>>
    >>> # Step 2: Create the binary model.
    >>> model = Model()
    >>> model.add_variable_vector('x', size, Vtype.BINARY)  # Notice "Vtype.BINARY"
    >>> model.set_objective_matrices(
    >>>     converter.converted_weights(),  # Notice usage of CONVERTED weights
    >>>     converter.converted_bias(),  # Notice usage of CONVERTED bias
    >>>     Target.MINIMIZE)
    >>>
    >>> # Step 3: Optimize the binary problem.
    >>> response = model.optimize()  # Add your usual arguments here.
    >>>
    >>> # Step 4: Convert the result back to bipolar.
    >>> for objective_value, result_vector in response.result_items():
    >>>     objective_value, result_vector = converter.convert_result(
    >>>         objective_value,
    >>>         result_vector,
    >>>         inplace=False)
    """

    def __init__(
            self,
            *,
            weights: Optional[np.ndarray]=None,
            bias: np.ndarray,
            inplace: bool,
            ) -> None:
        """
        Convert a bipolar problem definition to binary and construct a
        converter for its solution (back from binary into bipolar).

        Parameters
        ----------

        weights
            Weight matrix defining the bipolar problem.
            Refer to :meth:`titanq.Model.set_objective_matrices` for
            the format.
            If ``None``, an all-zero matrix is assumed.

        bias
            Bias vector defining the bipolar problem.
            Refer to :meth:`titanq.Model.set_objective_matrices` for
            the format.

        inplace
            If ``True``, modifies the weights and bias *in-place*,
            meaning the original object is modified and no copy is
            created.
            This is useful when dealing with very large objects.

        ⚠️ Warning
        ----------

        The created object can only convert a solution
        corresponding to the inputs used to construct it.
        If used with any other problem, the conversion will be
        incorrect.
        """

        if not inplace:
            if weights is not None:
                weights = weights.copy()
            bias = bias.copy()

        self._objective_value_offset: float = _convert_problem_in_place(weights, bias)
        self._binary_weights = weights
        self._binary_bias = bias

    def converted_weights(self) -> Optional[np.ndarray]:
        """
        Return the weights converted from a bipolar problem to a binary
        one.
        Only the values change. The shape and types remain the same.

        Return ``None`` if no weights were passed in.
        See :meth:`__init__`.

        If the weights were converted in-place, then the returned
        object is the modified one, not a copy.
        """
        return self._binary_weights

    def converted_bias(self) -> np.ndarray:
        """
        Return the bias converted from a bipolar problem to a binary
        one.
        Only the values change. The shape and types remain the same.

        If the bias was converted in-place, then the returned
        object is the modified one, not a copy.
        """
        return self._binary_bias

    def convert_result(
            self,
            objective_value: float,
            result_vector: np.ndarray,
            *,
            inplace: bool,
            ) -> Tuple[float, np.ndarray]:
        """
        Take a result from a binary computation and convert it back
        to be as if it was optimized as a bipolar problem.

        ⚠️ Warning
        ----------

        The conversion uses values computed from the original
        weights and bias.
        Do not use this object to convert a solution from a
        different problem than the one used to construct it.

        Parameters
        ----------

        objective_value
            The objective value returned from the binary optimization
            response.

        result_vector
            The result vector returned from the binary optimization
            response.

        inplace
            If ``True``, modifies the result vector *in-place*,
            meaning the original object is modified and no copy is
            created.
            This is useful when dealing with very large objects.

        Returns
        -------

        The objective value and result vector that would have been
        obtained if the problem had been optimized as bipolar.

        The objective value is converted to match the original weights
        and bias, before they were converted.

        The result vector is converted to bipolar values (``{-1, 1}``)
        instead of binary values (``{0, 1}``).
        """
        if not inplace:
            result_vector = result_vector.copy()

        converted_objective_value = objective_value + self._objective_value_offset
        _convert_binary_result_to_bipolar_in_place(result_vector)
        return converted_objective_value, result_vector


def _convert_problem_in_place(
        weights: Optional[np.ndarray],
        bias: np.ndarray,
        ) -> float:
    """
    Convert a bipolar problem into a binary one, *in-place*.

    Return the objective value offset, to reconvert the solution
    back from binary to bipolar later.
    """

    objective_value_offset = _compute_objective_value_offset(weights, bias)

    _convert_bias_in_place(weights, bias)

    if weights is not None:
        _convert_weights_in_place(weights)

    return objective_value_offset


def _convert_weights_in_place(weights: np.ndarray):
    weights *= 4


def _convert_bias_in_place(
        weights: Optional[np.ndarray],
        bias: np.ndarray):

    # bias_offset = weights dot ones(num_variables)
    bias_offset: float = 0  # Assume weights are all zeros if None.
    if weights is not None:
        num_variables = weights.shape[0]
        bias_offset = weights.dot(np.ones(num_variables))

    # bias = (bias - bias_offset) * 2
    bias -= bias_offset
    bias *= 2


def _convert_binary_result_to_bipolar_in_place(result_vector: np.ndarray):
    if not np.all((result_vector == 1) | (result_vector == 0)):
        raise ValueError("The result vector is not binary. Cannot convert to bipolar.")

    # result_vector * 2 - 1
    result_vector *= 2
    result_vector -= 1


def _compute_objective_value_offset(
        weights: Optional[np.ndarray],
        bias: np.ndarray,
        ) -> float:
    """
    Compute an offset to convert an objective value from a binary
    solution to match the original bipolar problem.

    Arguments must be the _BIPOLAR_ definition of the problem,
    _BEFORE_ conversion to binary.

    Returns an offset such that:
    ``bipolar_objective_value = binary_objective_value + offset``.
    """
    num_variables = bias.shape[0]
    ones_vector = np.ones(num_variables)

    weights_contribution = 0
    if weights is not None:
        weights_contribution = 0.5 * ones_vector.dot(weights.dot(ones_vector))

    return weights_contribution - bias.dot(ones_vector)



# Copyright (c) 2024, InfinityQ Technology, Inc.
import io
import json
import logging
import numpy as np
import numpy.typing as npt
import zipfile
from typing import Any, Dict, List, Optional, Tuple, Union

from .constraints import (
    Constraints,
    get_constraint_type,
    QuadConstraints,
)
from ..errors import (
    ConstraintAlreadySetError,
    MissingObjectiveError,
    MissingVariableError,
    ObjectiveAlreadySetError,
    OptimizeError,
)

from .manifest import Manifest
from .numpy_util import (
    is_ndarray_binary,
    is_upperbound_bigger_equal_lowerbound,
    ndarray_contains_nan_inf,
    reshape_to_2d,
    validate_cardinalities
)
from .objective import Objective, Target
from .optimize_response import OptimizeResponse
from .precision import Precision
from .math_object import Equation, Expression
from .variables import BinaryVariable, ContinuousVariable, IntegerVariable, Variable, Vtype

from .._client import api_model, Client
from .._storage import ManagedStorage, StorageClient, Ftype, to_bytes

log = logging.getLogger("TitanQ")


class Model:
    """
    Root object to define a problem to be optimized
    """

    def __init__(
        self,
        *,
        api_key: str = None,
        storage_client: StorageClient = None,
        base_server_url: str = "https://titanq.infinityq.io"
        ) -> None:
        """
        Initiate the model with a storage client. If the storage_client is missing, the storage will be managed by TitanQ.

        Notes
        -----
        The storage managed by TitanQ supports weight matrices with a size up to 10k only.

        Parameters
        ----------
        api_key
            TitanQ API key to access the service.
            If not set, it will use the environment variable ``TITANQ_API_KEY``
        storage_client
            Storage to choose in order to store some items.
        base_server_url
            TitanQ API server url, default set to ``https://titanq.infinityq.io``.

        Raises
        ------
        MissingTitanqApiKey
            If no API key is set and is also not set as an environment variable

        Examples
        --------
        With an S3 storage client
            >>> from titanq import Model, S3Storage
            >>> storage_client = S3Storage(
                access_key="{insert aws bucket access key here}",
                secret_key="{insert aws bucket secret key here}",
                bucket_name="{insert bucket name here}"
            )
            >>> model = Model(storage_client)

        Managed storage client
            >>> from titanq import Model, S3Storage
            >>> model = Model()
        """
        self._variables: List[Variable] = []
        self._objective: Objective = None
        self._constraints = Constraints()
        self._quad_constraints = QuadConstraints()

        self._titanq_client = Client(api_key, base_server_url)
        self._manifest: Manifest = Manifest()

        # the user chose a managed storage or left it as default
        if storage_client is None:
            storage_client = ManagedStorage(self._titanq_client)

        self._storage_client: StorageClient = storage_client


    def get_objective_matrices(self) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
        """
        Retrieve the weights and bias vector from the model's objective. Both will be None
        if not set.

        Return
        ------
        Weights matrix if not None and bias vector if not None
        """
        weights = self._objective.weights() if self._objective else None
        bias = self._objective.bias() if self._objective else None
        return (weights, bias)


    def get_constraints_weights_and_bounds(self) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
        """
        Retrieve the weights and bounds of all constraints from the model.

        Return
        ------
        Constraints weights if not None and constraints bounds if not None
        """
        return self._constraints.weights(), self._constraints.bounds()


    def get_quad_constraints_weights_and_bounds(self) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
        """
        Retrieve the quadratic constraints weights and bounds.

        Return
        ------
        Quadratic constraints weights if not None and quadratic constraints bounds if not None
        """
        return self._quad_constraints.weights(), self._constraints.bounds()


    def get_quad_constraints_linear_weights(self) -> Optional[np.ndarray]:
        """
        Retrieve the quadratic constraints linear weights.

        Return
        ------
        Quadratic constraints linear weights if not None
        """
        return self._quad_constraints.linear_weights()


    def add_variable_vector(
        self,
        name: str = '',
        size: int = 1,
        vtype: Vtype = Vtype.BINARY,
        variable_bounds:
            Optional[
                Union[
                    List[Tuple[int, int]],
                    List[Tuple[float, float]],
                ]
            ]=None,
    ) -> npt.NDArray[Any]:
        """
        Add a vector of variable to the model. Multiple variables vector can be added but with different names.

        Notes
        -----
        If Vtype is set to ``Vtype.INTEGER`` or ``Vtype.CONTINUOUS``, variable_bounds need to be set.

        Parameters
        ----------
        name
            The name given to this variable vector.
        size
            The size of the vector.
        vtype
            Type of the variables inside the vector.
        variable_bounds
            Lower and upper bounds for the variable vector. A list of tuples (can be either integers or continuous)

        Return
        ------
        variable
            The variable vector created.

        Raises
        ------
        MaximumVariableLimitError
            If the total size of variables exceed the limit.
        ValueError
            If the size of the vector is < 1

        Examples
        --------
        >>> from titanq import Model, Vtype
        >>> model.add_variable_vector('x', 3, Vtype.BINARY)
        >>> model.add_variable_vector('y', 2, Vtype.INTEGER, [[0, 5], [1, 6]])
        >>> model.add_variable_vector('z', 3, Vtype.CONTINUOUS, [[2.3, 4.6], [3.1, 5.3], [1.1, 4]])
        """

        if not variable_bounds:
            variable_bounds = []

        # validation
        if not self._constraints.is_empty() or not self._quad_constraints.is_empty():
            raise ConstraintAlreadySetError(
                "Cannot add additional variables once linear or quadratic " \
                "constraints have been defined")

        if self._objective is not None:
            raise ObjectiveAlreadySetError("Cannot add additional variable once objective have been defined")

        if vtype is Vtype.BINARY and variable_bounds:
            raise ValueError("variable_bounds is not supported with Vtype.BINARY")

        if size < 1:
            raise ValueError("Variable vector size cannot be less than 1")

        # create the variable
        variables = []
        if vtype is Vtype.BINARY:
            variables = [BinaryVariable(name, i, len(self._variables) + i) for i in range(size) ]
        elif vtype is Vtype.INTEGER:
            variables = [IntegerVariable(name, i, len(self._variables) + i, variable_bounds[i]) for i in range(size)]
        elif vtype is Vtype.CONTINUOUS:
            variables = [ContinuousVariable(name, i, len(self._variables)+i, variable_bounds[i]) for i in range(size)]
        else:
            raise NotImplementedError(f"Unsupported variable type: {vtype}")

        self._variables.extend(variables)
        log.debug(f"add variable name='{name}', type={str(vtype)}, size={size}.")

        return np.array([Expression(v) for v in variables])


    def set_objective_expression(self, expr: Expression, target=Target.MINIMIZE):
        """
        ℹ️ **This feature is experimental and may change.**

        Sets the objective function for the optimization problem using the given expression.

        This method processes the provided expression to extract the bias vector and weight matrix,
        and then sets these as the objective matrices for the optimization problem.

        Parameters
        ----------
        expr
            The expression defining the objective function.
        target
            The target of this objective matrix.

        Raises
        ------
        TypeError
            if the provided expression contains any invalid/unsupported input

        Examples
        --------
        >>> from titanq import Model, Vtype
        >>> x = model.add_variable_vector('x', 2, Vtype.BINARY)
        >>> y = model.add_variable_vector('y', 2, Vtype.BINARY)
        >>> expr = (np.array([3, 4]) * x + (x * y) - 5 * y)[0]
        >>> model.set_objective_expression(expr)
        """
        if not isinstance(expr, Expression):
            raise TypeError(f"The given objective expression is not of type {str(Expression)}.")

        constant, bias, weights = expr.split_into_component(len(self._variables))
        self.set_objective_matrices(weights, bias, target, constant)


    def add_constraint_from_expression(self, equation: Equation):
        """
        ℹ️ **This feature is experimental and may change.**

        Adds a constraint to the model using the given expression.

        This method processes the provided constraint expression to add it as a constraint to the
        optimization problem. Only linear constraints of the following types are supported:

        - `A == B`
        - `A <  B`
        - `A <= B`
        - `A >  B`
        - `A >= B`

        Constraints involving quadratic terms are not supported and will raise an error.

        Parameters
        ----------
        expression
            The constraint expression. This should be an instance of `Equation`.

        Raises
        ------
        ValueError
            If the provided expression contains quadratic terms.
        TypeError
            If the provided expression is of an invalid or unsupported type.

        Examples
        --------
        >>> from titanq import Model, Vtype
        >>> x = model.add_variable_vector('x', 2, Vtype.BINARY)
        >>> y = model.add_variable_vector('y', 2, Vtype.BINARY)
        >>> expr = sum(x+y) == 1
        >>> model.add_constraint_from_expression(expr)
        """
        if not isinstance(equation, Equation):
            raise TypeError(f"The given constraint equation is not of type {str(Equation)}.")

        n_variable = len(self._variables)
        mask, bounds = equation.generate_constraint(n_variable)

        self._constraints.add_constraint(
            variable_size=n_variable,
            constraint_weights=reshape_to_2d(mask),
            constraint_bounds=reshape_to_2d(bounds)
        )


    def set_objective_matrices(self, weights: Optional[np.ndarray], bias: np.ndarray, target=Target.MINIMIZE, constant_term: float=0.0):
        """
        Set the objective matrices for the model.

        Parameters
        ----------
        weights
            The quadratic objective matrix, **this matrix needs to be symmetrical**.
            A NumPy 2-D dense ndarray (must be float32).
            Weights matrix can be set to **None** if it is a linear problem with no quadratic elements.
        bias
            The linear constraint vector. A NumPy 1-D ndarray.
        target
            The target of this objective matrix.
        constant_term
            This parameter represents a fixed value added to the final computed objective value.

            Example: Suppose the objective function to minimize is 'f(x) = x^2 + 2x + 4' and 'x' is a binary variable.

            The optimal solution would be x = 0 resulting in an optimal objective function value of f(x)=4.
            However, the solver ignores constant terms and only seeks to minimize f(x) = x^2 + 2x.
            This would typically yield f(x)=0 by default with the optimal solution of x=0.
            constant_term should be set to 4 to yield an objective function value of f(x)=4.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        ObjectiveAlreadySetError
            If an objective has already been set in this model.
        ValueError
            If the weights shape or the bias shape does not fit the variables in the model.
            If the weights or bias data type is not float32.

        Examples
        --------
        >>> from titanq import Model, Target
        >>> edges = {0:[4,5,6,7], 1:[4,5,6,7], 2:[4,5,6,7], 3:[4,5,6,7], 4:[0,1,2,3], 5:[0,1,2,3], 6:[0,1,2,3], 7:[0,1,2,3]}
        >>> size = len(edges)
        >>> weights = np.zeros((size, size), dtype=np.float32)
        >>> for root, connections in edges.items():
        >>>     for c in connections:
        >>>         weights[root][c] = 1
        >>> # construct the bias vector (Uniform weighting across all nodes)
        >>> bias = np.asarray([0]*size, dtype=np.float32)
        >>> model.set_objective_matrices(weights, bias, Target.MINIMIZE)
        """
        if len(self._variables) == 0:
            raise MissingVariableError("Cannot set objective before adding a variable to the model.")

        if self._objective is not None:
            raise ObjectiveAlreadySetError("An objective has already have been set for this model.")

        log.debug(f"set objective matrix and bias vector.")

        self._objective = Objective(len(self._variables), weights, bias, target, constant_term)


    def add_set_partitioning_constraints_matrix(self, constraint_mask: np.ndarray):
        """
        Adds set partitioning constraints in matrix format to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 2-D dense ndarray (must be binary).
            The constraint_mask matrix of shape (M, N) where M the number of constraints and N is the number of variables.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of constraints exceed the limit.
        ConstraintSizeError
            If the constraint_mask shape does not fit the expected shape of this model.
        ValueError
            If the constraint_mask data type is not binary.

        Examples
        --------
        >>> constraint_mask = np.array([[1, 1, 1, 0, 1], [1, 1, 1, 1, 0]])
        >>> model.add_set_partitioning_constraints_matrix(constraint_mask)
        """
        if len(self._variables) == 0:
            raise MissingVariableError("Cannot set constraints before adding a variable to the model.")

        if constraint_mask.ndim == 1:
            raise ValueError(
                "Cannot use add_set_partitioning_constraints_matrix() function with a vector, " \
                "please use add_set_partitioning_constraint() instead")

        if not is_ndarray_binary(constraint_mask):
            raise ValueError(f"Cannot add a constraint if the values are not in binary.")

        self._constraints.add_constraint(
            variable_size=len(self._variables),
            constraint_weights=constraint_mask,
            constraint_bounds=np.ones((constraint_mask.shape[0],2))
        )


    def add_set_partitioning_constraint(self, constraint_mask: np.ndarray):
        """
        Adds set partitioning constraint vector to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 1-D dense ndarray (must be binary).
            The constraint_mask vector of shape (N,) where N is the number of variables.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of constraint exceed the limit.
        ConstraintSizeError
            If the constraint_mask shape does not fit the expected shape of this model.
        ValueError
            If the constraint_mask data type is not binary.

        Examples
        --------
        >>> constraint_mask = np.array([1, 1, 1, 0, 1])
        >>> model.add_set_partitioning_constraint(constraint_mask)
        """
        if constraint_mask.ndim > 1:
            raise ValueError(
                "Cannot use this add_set_partitioning_constraint() function with a matrix, " \
                "please use add_set_partitioning_constraints_matrix() instead")

        self.add_set_partitioning_constraints_matrix(reshape_to_2d(constraint_mask))


    def add_cardinality_constraints_matrix(self, constraint_mask: np.ndarray, cardinalities: np.ndarray):
        """
        Adds cardinality constraints in matrix format to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 2-D dense ndarray (must be binary).
            The constraint_mask matrix of shape (M, N) where M the number of constraints and N is the number of variables.
        cardinalities
            A NumPy 1-D ndarray (must be non-zero unsigned integer).
            The constraint_rhs vector of shape (M,) where M is the number of constraints.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of constraint exceed the limit.
        ConstraintSizeError
            If the constraint_mask shape or the constraint_rhs shape does not fit the expected shape of this model.
        ValueError
            If the constraint_mask is not binary or cardinalities data type are not unsigned integers.

        Examples
        --------
        >>> constraint_mask = np.array([[1, 1, 1, 0, 1], [1, 1, 1, 1, 0]])
        >>> cardinalities = np.array([3, 2])
        >>> model.add_cardinality_constraints_matrix(constraint_mask, cardinalities)
        """
        if len(self._variables) == 0:
            raise MissingVariableError("Cannot set constraints before adding a variable to the model.")

        if constraint_mask.ndim == 1:
            raise ValueError(
                "Cannot use add_cardinality_constraints_matrix() function with a vector, " \
                "please use add_cardinality_constraint() instead")

        if cardinalities.ndim != 1:
            raise ValueError(f"Cannot set constraints if cardinalities is not a NumPy 1-D dense ndarray")

        if not np.issubdtype(cardinalities.dtype, np.integer):
            raise ValueError("Found cardinalities data types not integer")

        if not np.all(cardinalities > 0):
            raise ValueError("Found cardinalities data types not unsigned integer")

        if not is_ndarray_binary(constraint_mask):
            raise ValueError(f"Cannot add a constraint if the values are not in binary.")

        if cardinalities.shape[0] != constraint_mask.shape[0]:
            raise ValueError(
                f"Cannot set constraints if cardinalities shape is not the same as the expected shape of this model." \
                f" Got cardinalities shape: {cardinalities.shape[1]}, constraint mask shape: {constraint_mask.size()}.")

        validate_cardinalities(constraint_mask, cardinalities)

        self._constraints.add_constraint(
            variable_size=len(self._variables),
            constraint_weights=constraint_mask,
            constraint_bounds=np.repeat(cardinalities, 2).reshape(-1, 2)
        )


    def add_cardinality_constraint(self, constraint_mask: np.ndarray, cardinality: int):
        """
        Adds cardinality constraint vector to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 1-D dense ndarray (must be binary).
            The constraint_mask vector of shape (N,) where N is the number of variables.
        cardinality
            The constraint_rhs cardinality.
            This value has to be a non-zero unsigned integer.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            If the number of constraints exceed the limit.
        ConstraintSizeError
            If the constraint_mask shape or the constraint_rhs shape does not fit
            the expected shape of this model.
        ValueError
            If the constraint_mask is not in binary or the cardinality is not an unsigned integer.

        Examples
        --------
        >>> constraint_mask = np.array([1, 1, 1, 0, 1])
        >>> cardinality = 3
        >>> model.add_cardinality_constraint(constraint_mask, cardinality)
        """
        if constraint_mask.ndim > 1:
            raise ValueError(
                "Cannot use add_cardinality_constraint() function with a matrix, " \
                "please use add_cardinality_constraints_matrix() instead")

        self.add_cardinality_constraints_matrix(reshape_to_2d(constraint_mask), np.full((1,), cardinality))


    def add_equality_constraints_matrix(self, constraint_mask: np.ndarray, limit: np.ndarray) -> None:
        """
        Adds an equality constraint matrix to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 2-D dense ndarray (float32).
            The constraint_mask vector of shape (M, N) where M the number of constraints and N is the number of variables.
        limit
            A NumPy 1-D array (float32).
            The limit vector of shape (M,) where M is the number of constraints.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of constraint exceed the limit.
        ValueError
            If the constraint_mask shape does not fit the expected shape of this model.
            If the constraint_mask or limit contains irregular format ('NaN' or 'inf').

        Examples
        --------
        >>> constraint_mask = np.array([[-3.51, 0, 0, 0], [10, 0, 0, 0]], dtype=np.float32)
        >>> limit = np.array([2, 10], dtype=np.float32)
        >>> model.add_equality_constraints_matrix(constraint_mask, limit)
        """
        if len(self._variables) == 0:
            raise MissingVariableError("Cannot set constraints before adding a variable to the model.")

        if constraint_mask.dtype != np.float32 or limit.dtype != np.float32:
            raise ValueError(f"Input parameters must be float32, got Constraint mask: {constraint_mask.dtype}, Limit: {limit.dtype}")

        if constraint_mask.ndim == 1:
            raise ValueError(
                "Cannot use add_equality_constraint_matrix() function with a vector, " \
                "please use add_equality_constraint() instead")

        if ndarray_contains_nan_inf(constraint_mask):
            raise ValueError("Constraint mask contains NaN or inf values")

        if limit.ndim != 1:
            raise ValueError("Limit must be a NumPy 1-D array")

        if ndarray_contains_nan_inf(limit):
            raise ValueError("Limit contains NaN or inf values")

        self._constraints.add_constraint(
            variable_size=len(self._variables),
            constraint_weights=constraint_mask,
            constraint_bounds=np.repeat(limit, 2).reshape(-1, 2)
        )


    def add_equality_constraint(self, constraint_mask: np.ndarray, limit: np.float32) -> None:
        """
        Adds an equality constraint vector to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 1-D dense ndarray (float32).
            The constraint_mask vector of shape (N,) where N is the number of variables.
        limit
            Limit value to the constraint mask.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of constraint exceed the limit.
        ValueError
            If the constraint_mask shape does not fit the expected shape of this model.
            If the constraint_mask or limit contains irregular format ('NaN' or 'inf').

        Examples
        --------
        >>> constraint_mask = np.array([1.05, -1.1], dtype=np.float32)
        >>> limit = -3.45
        >>> model.add_equality_constraint(constraint_mask, limit)
        """
        if constraint_mask.ndim > 1:
            raise ValueError(
                "Cannot use add_equality_constraint() function with a matrix, " \
                "please use add_equality_constraint_matrix() instead")

        self.add_equality_constraints_matrix(reshape_to_2d(constraint_mask), np.full((1,), limit, dtype=np.float32))


    def add_inequality_constraints_matrix(self, constraint_mask: np.ndarray, constraint_bounds: np.ndarray):
        """
        Adds inequality constraint matrix to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 2-D dense ndarray (float32).
            The constraint_mask vector of shape (M, N) where N is the number of variables.
        constraint_bounds
            A NumPy 2-D ndarray (float32).
            Vector of shape (M, 2) where M is the number of constraints.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of constraint exceed the limit.
        ValueError
            If the constraint_mask shape does not fit the expected shape of this model.
            If the constraint_mask contains irregular format ('NaN' or 'inf').
            If the lowerbound is equal or higher than its given upperbound.

        Examples
        --------
        >>> constraint_mask = np.array([[-3.51, 0], [10, 0]], dtype=np.float32)
        >>> constraint_bounds = np.array([[8, 9], [np.nan, 100_000]], dtype=np.float32)
        >>> model.add_inequality_constraints_matrix(constraint_mask, constraint_bounds)
        """
        if len(self._variables) == 0:
            raise MissingVariableError("Cannot set constraints before adding a variable to the model.")

        if constraint_mask.dtype != np.float32 or constraint_bounds.dtype != np.float32:
            raise ValueError(f"Input parameters must be float32, got Constraint mask: {constraint_mask.dtype}, Limit: {constraint_bounds.dtype}")

        if constraint_mask.ndim == 1:
            raise ValueError(
                "Cannot use add_inequality_constraint_matrix() function with a vector, " \
                "please use add_inequality_constraint() instead")

        if ndarray_contains_nan_inf(constraint_mask):
            raise ValueError("Constraint mask contains NaN or inf values.")

        if is_upperbound_bigger_equal_lowerbound(constraint_bounds):
            raise ValueError("Constraint bounds contains lowerbounds equal or higher than their upperbound.")

        self._constraints.add_constraint(
            variable_size=len(self._variables),
            constraint_weights=constraint_mask,
            constraint_bounds=constraint_bounds
        )


    def add_inequality_constraint(self, constraint_mask: np.ndarray, constraint_bounds: np.ndarray):
        """
        Adds inequality constraint vector to the model. At least one bound must be set.

        Parameters
        ----------
        constraint_mask
            A NumPy 1-D dense ndarray (float32).
            The constraint_mask vector of shape (N,) where N is the number of variables.
        constraint_bounds
            A NumPy 1-D ndarray (float32).
            Vector of shape (2,)

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of constraint exceed the limit.
        ValueError
            If the constraint_mask shape does not fit the expected shape of this model.
            If the constraint_mask contains irregular format ('NaN' or 'inf').
            If the lowerbound is equal or higher than the upperbound.

        Examples
        --------
        >>> constraint_mask = np.array([1.05, -1.1], dtype=np.float32)
        >>> constraint_bounds = np.array([1.0, np.nan], dtype=np.float32)
        >>> model.add_inequality_constraint(constraint_mask, constraint_bounds)
        """
        if constraint_mask.ndim > 1:
            raise ValueError(
                "Cannot use add_inequality_constraint() function with a matrix, " \
                "please use add_inequality_constraint_matrix() instead")

        self.add_inequality_constraints_matrix(reshape_to_2d(constraint_mask), reshape_to_2d(constraint_bounds))


    def add_quadratic_equality_constraint(
        self,
        constraint_mask: np.ndarray,
        limit: np.float32,
        constraint_linear_weights: Optional[np.ndarray] = None
    ) -> None:
        """
        ℹ️ **This feature is experimental and may change.**

        Adds an equality quadratic constraint to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 2-D dense ndarray (float32).
            The constraint_mask vector of shape (N, N) where N is the number of variables.
        limit
            Limit value to the constraint mask.
        constraint_linear_weights
            A NumPy 1-D dense ndarray (float32).
            The constraint_linear_weights vector of shape (N,) where N is the number of variables.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of quadratic constraint exceed the limit.
        ValueError
            If the constraint_mask shape does not fit the expected shape of this model.
            If the constraint_mask or limit contains irregular format ('NaN' or 'inf').

        Examples
        --------
        >>> constraint_mask = np.array([[0.1, 0.1], [0.1, 0.1]], dtype=np.float32)
        >>> limit = 1.0
        >>> constraint_linear_weights = np.array([0, 0.2], dtype=np.float32)
        >>> model.add_quadratic_equality_constraint(constraint_mask, limit, constraint_linear_weights)
        """
        self.add_quadratic_inequality_constraint(
            constraint_mask=constraint_mask,
            constraint_bounds=np.full((2,), limit, dtype=np.float32),
            constraint_linear_weights=constraint_linear_weights
        )


    def add_quadratic_inequality_constraint(
        self,
        constraint_mask: np.ndarray,
        constraint_bounds: np.ndarray,
        constraint_linear_weights: Optional[np.ndarray] = None
    ) -> None:
        """
        ℹ️ **This feature is experimental and may change.**

        Adds an inequality quadratic constraint to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 2-D dense ndarray (float32).
            The constraint_mask vector of shape (N, N) where N is the number of variables.
        constraint_bounds
            A NumPy 1-D ndarray (float32).
            Vector of shape (2,).
        constraint_linear_weights
            A NumPy 1-D dense ndarray (float32).
            The constraint_linear_weights vector of shape (N,) where N is the number of variables.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of quadratic constraint exceed the limit.
        ValueError
            If the constraint_mask shape does not fit the expected shape of this model.
            If the constraint_mask or limit contains irregular format ('NaN' or 'inf').
            If the constraint_bounds' shape is not (2,).

        Examples
        --------
        >>> constraint_mask = np.array([[1.05, -1.1], [0, 0]], dtype=np.float32)
        >>> constraint_bounds = np.array([np.nan, 10], dtype=np.float32)
        >>> constraint_linear_weights = np.array([4.0, 4.0], dtype=np.float32)
        >>> model.add_quadratic_inequality_constraint(constraint_mask, constraint_bounds, constraint_linear_weights)
        """
        if len(self._variables) == 0:
            raise MissingVariableError("Cannot set constraints before adding a variable to the model.")

        EXPECTED_BOUNDS_SHAPE = (2,)
        if constraint_bounds.shape != EXPECTED_BOUNDS_SHAPE:
            raise ValueError(
                "Invalid constraint_bounds shape: "
                + f"expected {EXPECTED_BOUNDS_SHAPE}, got {constraint_bounds.shape}")

        if constraint_bounds[0] > constraint_bounds[1]:
            raise ValueError("Input constraint_bounds lowerbound is higher than his upperbound.")

        if constraint_mask.dtype != np.float32:
            raise ValueError(f"Input parameters must be float32, got constraint_mask: {constraint_mask.dtype}")

        if ndarray_contains_nan_inf(constraint_mask):
            raise ValueError("Input parameter constraint_mask contains NaN or inf values")

        if constraint_linear_weights is not None:
            constraint_linear_weights = reshape_to_2d(constraint_linear_weights)

            if constraint_mask.dtype != np.float32:
                raise ValueError(f"Input parameters must be float32, got Constraint linear weights: {constraint_linear_weights.dtype}")

            if ndarray_contains_nan_inf(constraint_linear_weights):
                raise ValueError("Input parameter constraint_linear_weights contains NaN or inf values")

        self._quad_constraints.add_constraint(
            variable_size=len(self._variables),
            quad_constraint_weights=constraint_mask,
            quad_constraint_bounds=reshape_to_2d(constraint_bounds),
            quad_constraint_linear_weights=constraint_linear_weights
        )


    def optimize(
        self,
        *,
        beta: List[float] = [1, 0.5, 0.33, 0.25, 0.2, 0.16, 0.14, 0.125],
        coupling_mult: float = 0.5,
        timeout_in_secs: float = 10.0,
        num_chains: int = 8,
        num_engines: int = 1,
        penalty_scaling: float = None,
        precision: Precision = Precision.AUTO,
        num_buckets: int = 10
    ) -> OptimizeResponse:
        """
        Optimize this model. Issue a solve request and wait for it to complete.

        Notes
        -----
        All of the files used during this computation will be cleaned at the end.
        For more information on how to tunes those parameters, visit

        `The tuning guide <https://docs.titanq.infinityq.io/user-guide/parameter_tuning_guide>`_

        `TitanQ API documentation <https://docs.titanq.infinityq.io/>`_

        Parameters
        ----------
        beta
            Scales the problem by this factor (inverse of temperature). Beta values can then be
            adjusted to see if a better objective function value can be obtained.
            A lower beta allows for easier escape from local minima, while a higher beta
            is more likely to respect penalties and constraints.

            `Beta values tuning guide <https://docs.titanq.infinityq.io/user-guide/parameter_tuning_guide#beta>`_

            Range: List of [0, 20000]

            Recommended values: List of [0.004...2]

            NOTE: Beta values should be provided in descending order

            >>> import numpy as np
            >>> num_chains = 8
            >>> beta = (1/(np.linspace(2, 50, num_chains, dtype=np.float32))).tolist()

        coupling_mult
            Strength of parameter that keeps multiple logical copies of variables to have the
            same ground state solution. Heuristic to be tuned for a particular problem. Small values
            of this parameter will lead to an incorrect solution while large values will take a long
            time to converge to the correct solution.

            `coupling_mult tuning guide <https://docs.titanq.infinityq.io/user-guide/parameter_tuning_guide#coupling_mult>`_

            Range: [0, 100]

            Recommended values: [0.05...1.0]

        timeout_in_secs
            Maximum time (in second) the computation can take.

            `timeout_in_secs tuning guide <https://docs.titanq.infinityq.io/user-guide/parameter_tuning_guide#timeout_in_secs>`_

            Range: [0.1, 600]

            NOTE: Currently there is no other stop criteria. All computations will run up to the
            timeout value.

        num_chains
            Number of parallel chains running computation. Only the best result of all
            chains is returned.

            `num_chains tuning guide <https://docs.titanq.infinityq.io/user-guide/parameter_tuning_guide#num_chains--num_engines>`_

            Recommended values: [8, 16, 32]

            NOTE: ``num_chains`` * ``num_engines`` cannot exceed 512

        num_engines
            Number of independent batches of chains to run the computation. The best result
            of the batch of chains in each engine is returned.

            `num_engines tuning guide <https://docs.titanq.infinityq.io/user-guide/parameter_tuning_guide#num_chains--num_engines>`_

            Range: [1, 512]

            NOTE: ``num_chains`` * ``num_engines`` cannot exceed 512

        penalty_scaling
            Scaling factor for constraint penalty violations. Increasing this value results
            in stronger constraint enforcement at the cost of increasing the odds of becoming
            trapped in local minima.

            Range: penalty_scaling > 0

            NOTE: If None, a value will be inferred from the objective function of the problem.

        precision
            Some problems may need a higher precision implementation to converge properly such
            as when problem weights exhibit a high dynamic range. This flag allows this higher
            precision (but slightly slower speed) implementation to be used when ``Precision.HIGH``
            is set. Setting ``Precision.STANDARD`` uses a medium precision perfect for general use
            and offering the best speed/efficiency. The default setting ``Precision.AUTO`` will
            inspect the problem passed in and determine which of ``Precision.AUTO`` or
            ``Precision.STANDARD`` precision to use based on internal metrics.

        num_buckets
            The buckets strategy consists in dividing the range of values of a continuous variable
            into the specified number of buckets.

            `num_buckets tuning guide <https://docs.titanq.infinityq.io/user-guide/parameter_tuning_guide#num_buckets>`_

            Range: num_buckets >= 0

            Recommended values: [1, 2, 3, ..., 9, 10, 20, 30, ..., 90, 100, 200, 300, ..., 900, 1000, ...]

            NOTE: If the value is set to '0', the solver will not use the buckets strategy.

        Returns
        -------
        OptimizeResponse
            Optimized response data object

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MissingObjectiveError
            If no objective matrices have been added to the model.
        OptimizeError
            If an error was returned by the solver instead of a result,
            containing the optimize error message.

        Examples
        --------
        basic solve
            >>> response = model.optimize(timeout_in_secs=60)
        multiple engine
            >>> response = model.optimize(timeout_in_secs=60, num_engines=2)
        custom values
            >>> response = model.optimize(beta=[0.1], coupling_mult=0.75, num_chains=8)
        print values
            >>> print("-" * 15, "+", "-" * 26, sep="")
            >>> print("Ising energy   | Result vector")
            >>> print("-" * 15, "+", "-" * 26, sep="")
            >>> for ising_energy, result_vector in response.result_items():
            >>>     print(f"{ising_energy: <14f} | {result_vector}")
        """
        if len(self._variables) == 0:
            raise MissingVariableError("Cannot optimize before adding a variable to the model.")

        if self._objective is None:
            raise MissingObjectiveError("Cannot optimize before adding an objective to the model.")

        result, metrics = self._solve(beta, coupling_mult, timeout_in_secs, num_chains, num_engines, penalty_scaling, precision, num_buckets)

        return OptimizeResponse(self._variables, result, metrics)

    def _solve(
        self,
        beta: List[float],
        coupling_mult: float,
        timeout_in_secs: float,
        num_chains: int,
        num_engines: int,
        penalty_scaling: Optional[float],
        precision: Precision,
        num_buckets: int,
    ) -> Tuple[np.ndarray, Dict[str, Any]]:
        """
        Issue a solve request and wait for it to complete.

        Returns
        -------
        The result numpy array and the metric json object.
        """
        with self._storage_client as storage:
            storage.upload(Ftype.BIAS, to_bytes(self._objective.bias()))
            if self._objective.weights()is not None:
                storage.upload(Ftype.WEIGHTS, to_bytes(self._objective.weights()))
            if self._constraints.bounds() is not None and self._constraints.weights() is not None:
                storage.upload(Ftype.CONSTRAINT_BOUNDS, to_bytes(self._constraints.bounds()))
                storage.upload(Ftype.CONSTRAINT_WEIGHTS, to_bytes(self._constraints.weights()))
            storage.upload(Ftype.VARIABLE_BOUNDS, to_bytes(np.array([v.variable_bounds() for v in self._variables], dtype=np.float32)))
            if self._quad_constraints.weights() is not None and self._quad_constraints.bounds() is not None:
                storage.upload(Ftype.QUAD_CONSTRAINT_WEIGHTS, to_bytes(self._quad_constraints.weights()))
                storage.upload(Ftype.QUAD_CONSTRAINT_BOUNDS, to_bytes(self._quad_constraints.bounds()))
            if self._quad_constraints.linear_weights() is not None:
                storage.upload(Ftype.QUAD_CONSTRAINT_LINEAR_WEIGHTS, to_bytes(self._quad_constraints.linear_weights()))


            # api request
            request = api_model.SolveRequest(
                input=storage.input(),
                output=storage.output(),
                parameters=api_model.Parameters(
                    beta=beta,
                    coupling_mult=coupling_mult,
                    num_buckets=num_buckets,
                    num_chains=num_chains,
                    num_engines=num_engines,
                    penalty_scaling=penalty_scaling,
                    timeout_in_secs=timeout_in_secs,
                    variable_types=''.join(v.vtype()._api_str() for v in self._variables),
                    precision=str(precision),
                    constant_term=self._objective.constant_term()
                )
            )

            # adding constraint types
            if self._constraints.bounds() is not None and self._constraints.weights() is not None:
                for mask, bounds in zip(self._constraints.weights(), self._constraints.bounds()):
                    self._manifest.add_constraint_type(get_constraint_type(mask, bounds))

            # adding manifest flags to the request
            request.input.manifest = api_model.Manifest(
                has_set_partitioning_constraint=self._manifest.has_set_partitioning_constraint(),
                has_cardinality_constraint=self._manifest.has_cardinality_constraint(),
                has_equality_constraint=self._manifest.has_equality_constraint(),
                has_inequality_constraint=self._manifest.has_inequality_constraint()
            )

            solve_response = self._titanq_client.solve(request)

            archive_file_content = storage.wait_for_result_to_be_uploaded_and_download()
            metrics_content, result_content = _extract_result_files_or_raise(
                archive_file_content,
                solve_response.computation_id)

        log.debug("Optimization completed")
        return np.load(io.BytesIO(result_content)), json.loads(metrics_content)


def _extract_result_files_or_raise(
        archive_file_content: bytes,
        computation_id: str,
    ) -> Tuple[bytes, bytes]:
    """
    Take the raw content of a zip file returned by the solver, and
    return the unzipped contents of the metrics and the result files.

    If the zip file is invalid, or if it contains an error.json instead,
    raises an exception. The computation ID will always be included
    in the error message, so that we get it when users send their
    error message.
    """

    try:
        with zipfile.ZipFile(io.BytesIO(archive_file_content), 'r') as zip_file:
            try:
                metrics_content = zip_file.read("metrics.json")
                result_content = zip_file.read("result.npy")
            except (KeyError) as ex:
                try:
                    error_content = zip_file.read("error.json")
                    error_message = (
                        json.loads(error_content)["error"]
                        + f" (computation ID = {computation_id})")
                    raise OptimizeError(error_message)
                except (TypeError, KeyError, json.decoder.JSONDecodeError) as ex:
                    raise OptimizeError(
                        "Unexpected error in the solver, please contact InfinityQ support for more help" \
                        f" and provide the following computation ID: {computation_id}") from ex
    except zipfile.BadZipFile as ex:
        raise OptimizeError(
            "Failed to open the result zip file. Please contact InfinityQ support for more help" \
            f" and provide the following computation ID: {computation_id}") from ex

    return metrics_content, result_content



# Copyright (c) 2024, InfinityQ Technology, Inc.
import logging
import os
import requests

from pydantic import BaseModel, ValidationError
from typing import Any, Dict, Optional, Type
from urllib.parse import urljoin

import requests.adapters
import urllib3

from .model import CreditsResponse, SolveRequest, SolveResponse, TempStorageResponse
from ..errors import MissingTitanqApiKey, ServerError


log = logging.getLogger("TitanQ")

_QUEUED_STATUS = "Queued"
_TITANQ_API_VERSION = "v1"
_USER_AGENT_HEADER = 'User-Agent'

class Client:
    """
    TitanQ api client is a simple wrapper around TitanQ api to help interact with the
    service without the need to deal with http request
    """
    def __init__(self, api_key: Optional[str], base_server_url: str) -> None:
        api_key = api_key or os.getenv("TITANQ_API_KEY")
        if api_key is None:
            raise MissingTitanqApiKey(
                "No API key is provided. You can set your API key in the Model, "
                + "or you can set the environment variable TITANQ_API_KEY")
        self._server_url = base_server_url
        self._api_key = api_key


    def temp_storage(self) -> TempStorageResponse:
        """
        Query temporary storage url's

        :return: The temporary storage response

        :raises requests.exceptions.HTTPError: If an unexpected Error occur during request.
        """
        return self._do_http_request(f"{_TITANQ_API_VERSION}/temp_storage", response_type=TempStorageResponse)

    def credits(self) -> CreditsResponse:
        """
        Query Amount of credits remaining

        :return: The credit response.

        :raises requests.exceptions.HTTPError: If an unexpected Error occur during request.
        """
        return self._do_http_request(f"{_TITANQ_API_VERSION}/credits", response_type=CreditsResponse)

    def solve(self, request: SolveRequest) -> SolveResponse:
        """
        Issue a new solve request to the backend

        :param request: The solve request to issue to the solver

        :return: The response to the solve request (Not the response of the computation)

        :raises ServerError: If an unexpected Error occur during a solver request
        """
        log.debug(f"Issuing solve request to TitanQ server ({self._server_url}): {request}")
        response = self._do_http_request(f"{_TITANQ_API_VERSION}/solve", body=request, method='POST', response_type=SolveResponse)

        # something went wrong and the computation was not queued
        if response.status != _QUEUED_STATUS:
            log.error("An error occurred while issuing a solver request to the TitanQ server")
            raise ServerError(response.message)

        log.debug(f"Solve request response: {response}")
        return response


    def _do_http_request(
            self,
            path: str,
            *,
            headers: Dict[str, str] = {},
            body: BaseModel = None,
            method='GET',
            response_type: Type[BaseModel]
        ) -> Any:
        """
        Execute the actual http request to the TitanQ api while adding all defaults params

        :param headers: non-default header to the request.
        :param body: Body of the request.
        :param method: Which http method to use while performing the request.
        :param response_type: The object class that the json response will be cast to.

        :raise HTTPError: If the response cannot be created from the response type passed.

        :return: The response object created from the json response of the http request.
        """
        headers['authorization'] = self._api_key
        headers[_USER_AGENT_HEADER] = self._user_agent_string()
        url = urljoin(self._server_url, path)
        with requests.Session() as session:
            retries = urllib3.Retry(
                        total=3,
                        backoff_factor=0.5,
                        status_forcelist=[502, 503, 504, 495],
                        allowed_methods={"POST", "GET"},
                    )
            session.mount('https://', requests.adapters.HTTPAdapter(max_retries=retries))
            method = method.upper()
            if method=='GET':
                response = session.get(url, headers=headers)
            elif method=='POST':
                response = session.post(url, headers=headers, data=body.model_dump_json())
            else:
                raise NotImplementedError(f"http method: {method}")

            try:
                # create the response object from the response body
                return response_type.model_validate_json(response.content)
            except ValidationError:
                response.raise_for_status()
                raise

    def _user_agent_string(self) -> str:
        from titanq import __version__ as titanq_version # importing current module without cycle

        request_user_agent = requests.utils.default_headers().get(_USER_AGENT_HEADER, '')
        titanq_user_agent = f"TitanQ-sdk/{titanq_version} " + request_user_agent
        return titanq_user_agent.rstrip()